from abc import ABC, abstractmethod
from typing import List, Tuple, Any, Iterable, Union

from drepr.outputs.base_output_class import BaseOutputClass
from drepr.outputs.base_output_predicate import BaseOutputPredicate
from drepr.outputs.base_record import BaseRecord


class BaseSubsetOutputClass(BaseOutputClass):
    pass
